from babelfont import *
from babelfont.convertors import BaseConvertor
from fontTools.designspaceLib import DesignSpaceDocument
import ufoLib2
import uuid
import logging


log = logging.getLogger(__name__)


class Designspace(BaseConvertor):
    suffix = ".designspace"

    @classmethod
    def load(cls, convertor):
        self = cls()
        self.ds = DesignSpaceDocument.fromfile(convertor.filename)
        self.ds.loadSourceFonts(ufoLib2.Font.open)
        self.font = Font()
        return self._load()

    def _load(self):
        self._load_axes()

        for source in self.ds.sources:
            source._nfsf_master = self._load_master(source)
            self.font.masters.append(source._nfsf_master)

        for instance in self.ds.instances:
            self.font.instances.append(self._load_instance(instance))

        firstmaster = self.ds.sources[0].font
        self._load_metadata(firstmaster)
        glyphs_dict = self._load_glyphs(firstmaster)

        # Right, let's find all the layers. This will be messy.
        for source in self.ds.sources:
            for ufo_layer in source.font.layers:
                for g in source.font.glyphOrder:
                    if g not in glyphs_dict:
                        log.warn(
                            "Incompatible glyph set: %s appears in %s but is not in default"
                            % (g, source.filename)
                        )
                        continue
                    if g not in ufo_layer:
                        continue
                    glyphs_dict[g].layers.append(self._load_layer(source, ufo_layer, g))
        self._fixup_glyph_exported(self.ds.sources[0].font)
        return self.font

    def _fixup_glyph_exported(self, ufo):
        for glyph in ufo.lib.get("public.skipExportGlyphs", []):
            self.font.glyphs[glyph].exported = False

    def _load_glyphs(self, master):
        glyphs_dict = {}
        for g in master.glyphOrder:
            glyphs_dict[g] = self._load_glyph(master[g])
            self.font.glyphs.append(glyphs_dict[g])
        return glyphs_dict

    def _load_axes(self):
        for a in self.ds.axes:
            self.font.axes.append(
                Axis(
                    name=a.name,
                    tag=a.tag,
                    min=a.minimum,
                    max=a.maximum,
                    default=a.default,
                    map=a.map,
                )
            )

    def _load_master(self, source):
        i = source.font.info
        master = Master(
            name=source.name,
            id=(source.name or uuid.uuid1()),
        )
        for metric in Master.CORE_METRICS:
            master.metrics[metric] = getattr(i, metric)
        _axis_name_to_id = {a.name.get_default(): a.tag for a in self.font.axes}
        # italic angle
        # names XXX
        # guidelines
        master.guides = [self._load_guide(g) for g in (i.guidelines or [])]
        master.location = {_axis_name_to_id[k]: v for k, v in source.location.items()}
        master.font = self.font
        master.kerning = self._load_kerning(source)
        self._load_groups(source.name, source.font.groups)
        assert master.valid
        return master

    def _load_groups(self, sourcename, groups):
        for name, value in groups.items():
            if (
                name in self.font.features.namedClasses
                and self.font.features.namedClasses[name] != value
            ):
                log.warn(
                    "Inconsistent definition of glyph class @%s found in %s"
                 % (name, sourcename))
            self.font.features.namedClasses[name] = value

    def _load_guide(self, ufo_guide):
        return Guide(
            pos=[ufo_guide.x, ufo_guide.y, ufo_guide.angle],
            name=ufo_guide.name,
            color=ufo_guide.color,
        )

    def _load_kerning(self, source):
        kerning = {}
        for (l, r), value in source.font.kerning.items():
            if l.startswith("public.kern"):
                l = "@" + l
            if r.startswith("public.kern"):
                r = "@" + r
            kerning[(l, r)] = value
        return kerning

    def _load_glyph(self, ufo_glyph):
        cp = ufo_glyph.unicodes or [ufo_glyph.unicode]
        lib = self.ds.sources[0].font.lib
        category = lib.get("public.openTypeCategories", {}).get(ufo_glyph.name, "base")
        g = Glyph(name=ufo_glyph.name, codepoints=cp, category=category)
        if "public.postscriptNames" in lib and g.name in lib["public.postscriptNames"]:
            g.production_name = lib["public.postscriptNames"][g.name]
        return g

    def _load_layer(self, source, ufo_layer, glyphname):
        ufo_glyph = ufo_layer[glyphname]
        width = ufo_glyph.width
        l = Layer(width=width, id=uuid.uuid1())
        l._master = source._nfsf_master.id
        l._font = self.font
        for contour in ufo_glyph:
            l.shapes.append(self._load_contour(contour))
        for component in ufo_glyph.components:
            l.shapes.append(self._load_component(component))
        for anchor in ufo_glyph.anchors:
            l.anchors.append(self._load_anchor(anchor))
        assert l.valid
        return l

    def _load_component(self, shape):
        c = Shape(ref=shape.baseGlyph, transform=shape.transformation)
        return c

    def _load_anchor(self, anchor):
        return Anchor(name=anchor.name, x=int(anchor.x), y=int(anchor.y))

    def _load_contour(self, contour):
        shape = Shape()
        shape.nodes = []
        for p in contour:
            segtype = p.segmentType
            if p.segmentType == "move":
                segtype = "line"
            ourtype = Node._from_pen_type[segtype]
            if p.smooth:
                ourtype = ourtype + "s"
            shape.nodes.append(Node(p.x, p.y, ourtype))
        return shape

    def _load_instance(self, ufo_instance):
        _axis_tag = {axis.name.get_default(): axis.tag for axis in self.font.axes}
        location = {_axis_tag[k]: v for k, v in ufo_instance.location.items()}
        instance = Instance(name=ufo_instance.name, location=location)
        return instance

    names_dict = {
        "designer": "openTypeNameDesigner",
        "designerURL": "openTypeNameDesignerURL",
        "manufacturer": "openTypeNameManufacturer",
        "manufacturerURL": "openTypeNameManufacturerURL",
        "license": "openTypeNameLicense",
        "licenseURL": "openTypeNameLicenseURL",
        "version": "openTypeNameVersion",
        "uniqueID": "openTypeNameUniqueID",
        "description": "openTypeNameDescription",
        "compatibleFullName": "openTypeNameCompatibleFullName",
        "sampleText": "openTypeNameSampleText",
        "WWSFamilyName": "openTypeNameWWSFamilyName",
        "WWSSubfamilyName": "openTypeNameWWSSubfamilyName",
        "copyright": "copyright",
        "styleMapFamilyName": "styleMapFamilyName",
        "familyName": "familyName",
        "trademark": "trademark",
    }

    def _load_metadata(self, ufo):
        firstfontinfo = ufo.info
        self.font.upm = firstfontinfo.unitsPerEm
        self.font.version = (firstfontinfo.versionMajor, firstfontinfo.versionMinor)
        self.font.note = firstfontinfo.note
        for ours, theirs in self.names_dict:
            their_value = firstfontinfo.getattr(theirs)
            if their_value:
                getattr(self.font.names, ours).set_default(their_value)



from datetime import datetime
from babelfont import *
import openstep_plist
from fontTools.misc.transform import Transform
from fontFeatures.feaLib import FeaParser
from babelfont.convertors import BaseConvertor
import re
import math
import uuid
from collections import OrderedDict
from glyphsLib.glyphdata import get_glyph
from fontTools.misc.filenames import userNameToFileName
import os


opentype_custom_parameters = {
    "typoAscender": ("OS/2", "sTypoAscender"),
    "typoDescender": ("OS/2", "sTypoDescender"),
    "typoLineGap": ("OS/2", "sTypoLineGap"),
    "winAscent": ("OS/2", "usWinAscent"),
    "winDescent": ("OS/2", "usWinDescent"),
    "hheaAscender": ("hhea", "ascent"),
    "hheaDescender": ("hhea", "descent"),
    "hheaLineGap": ("hhea", "lineGap"),
    "underlinePosition": ("post", "underlinePosition"),
    "underlineThickness": ("post", "underlineThickness"),
}

_rename_metrics = {"x-height": "xHeight", "cap height": "capHeight"}
_reverse_rename_metrics = {v: k for k, v in _rename_metrics.items()}


def _glyphs_metrics_to_ours(k):
    return _rename_metrics.get(k, k)


def _our_metrics_to_glyphs(k):
    return _reverse_rename_metrics.get(k, k)


class GSObject(BaseConvertor):
    @classmethod
    def can_load(cls, convertor):
        return "gsfont" in convertor.scratch

    def _load(self):
        self.gsfont = self.scratch["gsfont"]
        self.customParameters = {}
        for param in self.gsfont.customParameters:
            self.customParameters[param.name] = param.value
        self._load_axes()
        self._load_kern_groups(self.gsfont.glyphs)

        for gmaster in self.gsfont.masters:
            self.font.masters.append(self._load_master(gmaster))
        self._fixup_axes()

        for gglyph in self.gsfont.glyphs:
            g = self._load_glyph(gglyph)
            self.font.glyphs.append(g)

        for ginstance in self.gsfont.instances:
            self.font.instances.append(self._load_instance(ginstance))

        self._fixup_axis_mappings()

        self._load_metadata()
        self._load_features()
        return self.font

    def _fixup_axes(self):
        for master in self.font.masters:
            for axis in self.font.axes:
                thisLoc = master.location[axis.tag]
                if axis.min is None or thisLoc < axis.min:
                    axis.min = thisLoc
                if master.id == self._default_master_id():
                    axis.default = master.location[axis.tag]
                if axis.max is None or thisLoc > axis.max:
                    axis.max = thisLoc

    def _fixup_axis_mappings(self):
        for axis in self.font.axes:
            if not axis.map:
                continue
            axis.map = list(sorted(set(axis.map)))
            axis.min, axis.max, axis.default = (
                axis.map_backward(axis.min),
                axis.map_backward(axis.max),
                axis.map_backward(axis.default),
            )

    def _custom_parameter(self, thing, name):
        for param in (thing.customParameters or []):
            if param.name == name:
                return param.value
        return None

    def _default_master_id(self):
        # The default master in glyphs is either the first master or the
        # one selected by the Variable Font Origin custom parameter
        return (
            self._custom_parameter(self.gsfont, "Variable Font Origin")
            or self.gsfont.masters[0].id
        )

    def _load_axes(self):
        self.axis_name_map = {}
        for gaxis in self.gsfont.axes:
            axis = Axis(name=gaxis.name, tag=gaxis.axisTag)
            self.axis_name_map[gaxis.name] = axis
            _maybesetformatspecific(axis, gaxis, "hidden")
            self.font.axes.append(axis)

    def _load_master(self, gmaster):
        location = gmaster.axes
        metrics = self.gsfont.metrics()
        master = Master(name=gmaster.name, id=gmaster.id, font=self.font)
        metric_types = [m.name or m.typeName() for m in metrics]
        for k, v in zip(metric_types, gmaster.metrics()):
            master.metrics[_glyphs_metrics_to_ours(k)] = v.position
            if v.overshoot:
                master.metrics["%s overshoot" % _glyphs_metrics_to_ours(k)] = v.overshoot

        master.location = {k.tag: v for k, v in zip(self.font.axes, location)}
        master.guides = [self._load_guide(x) for x in gmaster.guides]
        kerntable = self.gsfont.kerning.get(master.id, {})
        master.kerning = self._load_kerning(kerntable)
        # XXX support RTL kerning etc.

        # _maybesetformatspecific(master, gmaster, "customParameters") # XXX
        _maybesetformatspecific(master, gmaster, "iconName")
        _maybesetformatspecific(master, gmaster, "numberValues")
        _maybesetformatspecific(master, gmaster, "stemValues")
        _maybesetformatspecific(master, gmaster, "properties")
        _maybesetformatspecific(master, gmaster, "userData")
        _maybesetformatspecific(master, gmaster, "visible")
        assert master.valid
        return master

    def _get_codepoint(self, gglyph):
        if gglyph.unicodes:
            return [int(x,16) for x in gglyph.unicodes]
        return []

    def _load_glyph(self, gglyph):
        name = gglyph.name
        c = gglyph.category
        sc = gglyph.subCategory
        if sc == "Ligature":
            category = "ligature"
        if c == "Mark":
            category = "mark"
        else:
            category = "base"
        cp = self._get_codepoint(gglyph)
        g = Glyph(
            name=name,
            codepoints=cp,
            category=category,
            exported=gglyph.export,
        )
        g.production_name = gglyph.productionName or get_glyph(name).production_name
        for entry in [
            "case",
            "category",
            "subCategory",
            "color",
            "direction",
            "locked",
            "partsSettings",
            "script",
            "tags",
        ]:
            _maybesetformatspecific(g, gglyph, entry)

        for layer in gglyph.layers:
            g.layers.extend(self._load_layer(layer))

        return g

    def _load_layer(self, layer, width=None):
        if width is None:
            width = layer.width
        l = Layer(width=width, id=str(layer.layerId), _font=self.font)
        l.name = layer.name
        if [x for x in self.font.masters if x.id == l.id]:
            l._master = l.id
        else:
            l._master = layer.associatedMasterId
        l.guides = [self._load_guide(x) for x in layer.guides]
        l.shapes = []
        for shape in layer.shapes:
            l.shapes.append(self._load_shape(shape))
        for anchor in layer.anchors:
            l.anchors.append(self._load_anchor(anchor))

        _maybesetformatspecific(l, layer, "hints")
        _maybesetformatspecific(l, layer, "partSelection")
        _maybesetformatspecific(l, layer, "visible")
        _maybesetformatspecific(l, layer, "attr")
        returns = [l]
        if "Background" not in str(type(layer)) and layer.background:
            (background,) = self._load_layer(layer.background, width=l.width)
            # If it doesn't have an ID, we need to generate one
            background.id = str(uuid.uuid1())
            # XXX
            # For some INSANE REASON orjson is serializing this to `True` in json,
            # not `true`.
            #background.isBackground = True
            del(background._master)

            l._background = background.id
            returns.append(background)
        # TODO backgroundImage, metricTop/Bottom/etc, vertOrigin, vertWidth.
        for r in returns:
            assert r.valid
        return returns

    def _load_guide(self, gguide):
        pos = gguide.position
        return Guide(pos=[int(pos[0]), int(pos[1]), int(gguide.angle)])

    def _load_anchor(self, ganchor):
        pos = ganchor.position
        return Anchor(name=ganchor.name, x=float(pos[0]), y=float(pos[1]))

    def _load_instance(self, ginstance):
        if ginstance.axes:
            location = ginstance.axes
            instance_location = {k.tag: v for k, v in zip(self.font.axes, location)}
        elif ginstance.instanceInterpolations:
            # All right then.
            instance_location = {k.tag: 0 for k in self.font.axes}
            for mId, factor in ginstance.instanceInterpolations.items():
                master_loc = self.font.master(mId).location
                for k in self.font.axes:
                    instance_location[k.tag] += master_loc[k.tag] * factor
        else:
            raise ValueError("Need to Synthesize location")
        i = Instance(
            name=ginstance.name,
            styleName=ginstance.name,
            location=instance_location,
        )
        c = self._custom_parameter(ginstance, "Axis Location") or []
        for loc in c:
            ax = self.axis_name_map[loc["Axis"]]
            if not ax.map:
                ax.map = []
            ax.map.append(
                (
                    int(loc["Location"]),
                    instance_location[ax.tag],
                )
            )

        # _maybesetformatspecific(i, ginstance, "customParameters") # XXX
        return i

    def _load_component(self, shape):
        glyphname = shape.componentName
        transform = shape.transform
        c = Shape(ref=glyphname)
        c.transform = transform
        for entry in [
            "alignment",
            "anchorTo",
            "attr",
            "locked",
            "orientation",
            "piece",
            "userData",
        ]:
            _maybesetformatspecific(c, shape, entry)
        return c

    def _load_kerning(self, kerndict):
        return {
            (l, r): value
            for l, level2 in kerndict.items()
            for r, value in level2.items()
        }

    def _load_metadata(self):
        self.font.upm = self.gsfont.upm
        self.font.version = (self.gsfont.versionMajor, self.gsfont.versionMinor)
        self.font.names.familyName.set_default(self.gsfont.familyName)

        # This is very glyphs 3
        props = {}
        for prop in self.gsfont.properties:
            if hasattr(prop, "value"):
                props[prop.key] = prop.value
            else:
                props[prop.key] = {p.languageTag: p.value for p in prop.values}

        if props:
            interestingProps = {
                "copyrights": "copyright",
                "designer": "designer",
                "designerURL": "designerURL",
            }  # Etc
            for glyphsname, attrname in interestingProps.items():
                thing = props.get(glyphsname, "")
                if hasattr(thing, dict):
                    getattr(self.font.names, attrname).copy_in(thing)
                else:
                    getattr(self.font.names, attrname).set_default(thing)
            # Do other properties here

        # Any customparameters in the default master which look like
        # custom OT values need to move there.
        cp = self.font.default_master._formatspecific.get("com.glyphsapp", {}).get(
            "customParameters", {}
        )
        for param in cp:
            ot_param = opentype_custom_parameters.get(param.name)
            if not ot_param:
                continue
            self.font.customOpenTypeValues.append(
                OTValue(ot_param[0], ot_param[1], param.value)
            )

        self.font.note = self.gsfont.note
        self.font.date = self.gsfont.date

        _maybesetformatspecific(self.font, self.gsfont, ".appVersion")
        _maybesetformatspecific(self.font, self.gsfont, ".formatVersion")
        _maybesetformatspecific(self.font, self.gsfont, "DisplayStrings")
        _maybesetformatspecific(self.font, self.gsfont, "customParameters")
        _maybesetformatspecific(self.font, self.gsfont, "settings")
        _maybesetformatspecific(self.font, self.gsfont, "numbers")
        _maybesetformatspecific(self.font, self.gsfont, "stems")
        _maybesetformatspecific(self.font, self.gsfont, "userData")
        _maybesetformatspecific(self.font, self.gsfont, "metrics")

    def _load_features(self):
        for c in self.gsfont.classes:
            self.font.features.namedClasses[c.name] = tuple(c.code.split())

        featurefile = ""
        for f in self.gsfont.featurePrefixes:
            featurefile = featurefile + f.code
        for f in self.gsfont.features:
            tag = f.name
            feacode = "feature %s { %s\n} %s;" % (tag, f.code, tag)
            featurefile = featurefile + feacode

        try:
            feaparser = FeaParser(featurefile)
            ast = feaparser.parser.ast
            for name, members in self.font.features.namedClasses.items():
                glyphclass = ast.GlyphClassDefinition(
                    name, ast.GlyphClass([m for m in members])
                )
                feaparser.parser.glyphclasses_.define(name, glyphclass)
            feaparser.parse()
        except Exception as e:
            pass

    def _load_shape(self, shape):
        if hasattr(shape, "nodes"):  # It's a path
            return self._load_path(shape)
        else:
            return self._load_component(shape)

    def _load_path(self, path):
        shape = Shape()
        shape.nodes = [Node(n.position.x, n.position.y, n.type[0]) for n in path.nodes]
        shape.closed = path.closed
        return shape

    def _load_kern_groups(self, glyphs):
        kerngroups = {}
        for g in glyphs:
            if g.leftKerningGroup:
                kerngroups.setdefault("MMK_L_" + g.leftKerningGroup, []).append(
                    g.name
                )
            if g.rightKerningGroup:
                kerngroups.setdefault("MMK_R_" + g.rightKerningGroup, []).append(
                    g.name
                )
        for k, v in kerngroups.items():
            self.font.features.namedClasses[k] = tuple(v)


def _maybesetformatspecific(item, glyphs, key):
    if hasattr(glyphs, key):
        if "com.glyphsapp" not in item._formatspecific:
            item._formatspecific["com.glyphsapp"] = {}
        value = getattr(glyphs, key) # XXX
        try:
            orjson.dumps(value)
            item._formatspecific["com.glyphsapp"][key] = value
        except Exception as e:
            print("%s.%s cannot be serialized!" % (glyphs, key))


def _moveformatspecific(item):
    rv = {}
    if "com.glyphsapp" in item._formatspecific:
        rv = {**item._formatspecific.get("com.glyphsapp", {})}
    return rv


def _copyattrs(src, dst, attrs):
    for a in attrs:
        v = getattr(src, a)
        if isinstance(v, I18NDictionary):
            v = v.get_default()
        if v:
            dst[a] = v



from dataclasses import dataclass, field
from .BaseObject import BaseObject
from .Layer import Layer
from fontTools.misc.filenames import userNameToFileName
import os


class GlyphList(dict):
    def append(self, thing):
        self[thing.name] = thing

    def write(self, stream, indent):
        stream.write(b"[")
        for ix, l in enumerate(self):
            stream.write(b"\n")
            stream.write(b"  " * (indent + 2))
            l.write(stream, indent + 1)
            if ix < len(self) - 1:
                stream.write(b", ")
            else:
                stream.write(b"\n")
        stream.write(b"]")

    def __iter__(self):
        self._n = 0
        self._values = list(self.values())
        return self

    def __next__(self):
        if self._n < len(self._values):
            result = self._values[self._n]
            self._n += 1
            return result
        else:
            raise StopIteration

@dataclass
class _GlyphFields:
    name: str
    production_name: str = None
    category: str = "base"
    codepoints: [int] = field(default_factory=list)
    layers: [Layer] = field(default_factory=list, repr=False, metadata={"skip_serialize": True})
    exported: bool = field(default=True, metadata={"serialize_if_false": True})
    direction: str = field(default="LTR", repr=False)


@dataclass
class Glyph(BaseObject, _GlyphFields):

    _write_one_line = True

    @property
    def nfsf_filename(self):
        return os.path.join("glyphs", (userNameToFileName(self.name) + ".nfsglyph"))




from babelfont import *
from babelfont.convertors.designspace import Designspace
import ufoLib2


class UFO(Designspace):
    suffix = ".ufo"

    @classmethod
    def load(cls, convertor):
        self = cls()
        self.ufo = ufoLib2.Font(convertor.filename)
        self.font = Font()
        return self._load()

    def _load(self):
        self.font.masters = [self._load_master(self.ufo)]
        self._load_metadata(self.ufo)



from datetime import datetime
from babelfont import *
from fontTools.fontBuilder import FontBuilder
from fontTools.pens.ttGlyphPen import TTGlyphPen
from fontTools.pens.recordingPen import RecordingPen
from cu2qu.ufo import glyphs_to_quadratic
from fontTools.misc.timeTools import epoch_diff, timestampSinceEpoch
from fontTools.ttLib.ttFont import _TTGlyphGlyf, _TTGlyphSet
from fontTools.ttLib.tables.TupleVariation import TupleVariation
from babelfont.fontFilters.featureWriters import build_all_features
from fontTools.ttLib import TTFont
from fontTools.ttLib.tables._g_l_y_f import GlyphCoordinates
from fontTools.varLib.iup import iup_delta_optimize
from fontTools.misc.fixedTools import otRound
import uuid
from itertools import chain


def _categorize_glyph(font,glyphname):
    if "GDEF" not in font: return None
    classdefs = font["GDEF"].table.GlyphClassDef.classDefs
    if glyphname not in classdefs:
        return None
    if classdefs[glyphname] == 1: return "base"
    if classdefs[glyphname] == 2: return "ligature"
    if classdefs[glyphname] == 3: return "mark"
    if classdefs[glyphname] == 4: return "component"

class TrueType(BaseConvertor):
    suffix = ".ttf"

    def _decompose_mixed_layer(self, layer, exportable):
        if (layer.paths and layer.components) or any(
            c.ref not in exportable for c in layer.components
        ):
            layer.decompose()

    def _load(self):
        self.tt = TTFont(self.filename)
        self._load_fvar()
        self._load_head()
        self._load_masters()
        self._load_names()
        self._load_glyphs()
        return self.font

    def _load_fvar(self):
        avar = self.tt.get("avar")
        if "fvar" in self.tt:
            for axis in self.tt["fvar"].axes:
                name = self.tt["name"].getDebugName(axis.axisNameID)  # XXX multilingual
                bb_axis = Axis(
                    tag=axis.axisTag,
                    min=axis.minValue,
                    max=axis.maxValue,
                    default=axis.defaultValue,
                    name=name,
                )
                self.font.axes.append(bb_axis)
                if avar:
                    segs = avar.segments[axis.axisTag]
                    mapping = {
                        bb_axis.denormalize_value(k): bb_axis.denormalize_value(v)
                        for k, v in segs.items()
                    }
                    bb_axis.map = mapping
            for instance in self.tt["fvar"].instances:
                self.font.instances.append(
                    Instance(
                        name=self.tt["name"].getDebugName(instance.subfamilyNameID),
                        location=instance.coordinates,
                    )
                )

    def _load_masters(self):
        m = Master(location={},name="Default", id=str(uuid.uuid1()) )
        # Metrics
        m.metrics = {
            "xHeight": self.tt["OS/2"].sxHeight,
            "capHeight": self.tt["OS/2"].sCapHeight,
            "ascender": self.tt["hhea"].ascender,
            "descender": self.tt["hhea"].descender
        }
        m.font = self.font
        self.font.masters = [m]
        if "fvar" in self.tt:
            m.location = {axis.tag: axis.default for axis in self.font.axes}
            all_masters = [frozenset(x.axes.items()) for x in chain(*self.tt["gvar"].variations.values())]
            all_masters = [{k:v[1] for k,v in dict(m1).items()} for m1 in all_masters]
            # Now denormalize.
            # XXX
            pass

    def _load_head(self):
        head = self.tt["head"]
        self.font.upm = head.unitsPerEm
        minor = head.fontRevision % 1
        while minor - int(minor) > 1e-4:
            minor *= 10
        self.font.version = (int(head.fontRevision), int(minor))
        self.font.date = datetime.fromtimestamp(self.tt["head"].created + epoch_diff)

    def _load_names(self):
        names = self.tt["name"]
        # XXX

    def _load_glyphs(self):
        mapping = self.tt["cmap"].buildReversed()
        glyphs_dict = {}
        for glyph in self.tt.getGlyphOrder():
            category = _categorize_glyph(self.tt, glyph) or "base"
            glyphs_dict[glyph] = Glyph(name=glyph, codepoints=list(mapping.get(glyph, [])), category=category)
            self.font.glyphs.append(glyphs_dict[glyph])
            glyphs_dict[glyph].layers = self._load_layers(glyph)
        return glyphs_dict

    def _load_layers(self, g):
        ttglyph = self.tt.getGlyphSet()[g]._glyph  # _TTGlyphGlyf object
        width = self.tt["hmtx"][g][0]
        # leftMargin = self.tt["hmtx"][g][1]
        layer = Layer(width=width, id=str(uuid.uuid1()) )
        layer._master = self.font.masters[0].id
        layer._font = self.font
        for i in range(0, max(ttglyph.numberOfContours, 0)):
            layer.shapes.append(self._load_contour(ttglyph, i))
        if hasattr(ttglyph, "components"):
            for c in ttglyph.components:
                comp = self._load_component(c)
                layer.shapes.append(comp)
        return [layer]


    def _load_contour(self, ttglyph, index):
        shape = Shape()
        shape.nodes = []
        endPt = ttglyph.endPtsOfContours[index]
        if index > 0:
            startPt = ttglyph.endPtsOfContours[index - 1] + 1
        else:
            startPt = 0
        points = []
        for j in range(startPt, endPt + 1):
            coords = (ttglyph.coordinates[j][0], ttglyph.coordinates[j][1])
            flags = ttglyph.flags[j] == 1
            t = "o"
            if flags == 1:
                if (j == startPt and ttglyph.flags[endPt] == 1) or (
                    j != startPt and points[-1].type != "o"
                ):
                    t = "l"
                else:
                    t = "q"
            else:
                if len(points) > 1 and points[-1].type == "o":
                    # Double offcurve. Insert implicit oncurve.
                    prevpoint = points[-1]
                    intermediate = Node(
                        x = (coords[0] + prevpoint.x) / 2,
                        y = (coords[1] + prevpoint.y) / 2,
                        type = "q"
                    )
                    points.append(intermediate)
            p = Node(x = coords[0], y=coords[1], type=t)
            points.append(p)
        shape.nodes = points
        return shape

    def _load_component(self, c):
        baseGlyph, transformation = c.getComponentInfo()
        component = Shape(ref=baseGlyph, transform=transformation)
        return component

    def _save(self):
        f = self.font
        fb = FontBuilder(f.upm, isTTF=True)

        metrics = {}
        all_outlines = {}

        # Find all exportable glyphs
        exportable = [k for k, v in f.glyphs.items() if v.exported]

        fb.setupGlyphOrder(exportable)
        fb.setupCharacterMap(
            {k: v for k, v in f.unicode_map.items() if v in exportable}
        )

        for g in exportable:
            all_outlines[g] = []
            layer = f.default_master.get_glyph_layer(g)
            metrics[g] = (layer.width, layer.lsb)

        fb.setupHorizontalMetrics(metrics)

        for m in f.masters:
            glyf = {}
            m.ttglyphset = _TTGlyphSet(fb.font, glyf, _TTGlyphGlyf)

        done = {}

        def do_a_glyph(g):
            if g in done:
                return
            layer = f.default_master.get_glyph_layer(g)
            self._decompose_mixed_layer(layer, exportable)
            for c in layer.components:
                do_a_glyph(c.ref)

            for m in f.masters:
                layer = m.get_glyph_layer(g)
                self._decompose_mixed_layer(layer, exportable)
                all_outlines[g].append(layer)
            try:
                glyphs_to_quadratic(all_outlines[g], reverse_direction=True)
                for ix, m in enumerate(f.masters):
                    layer = m.get_glyph_layer(g)
                    pen = TTGlyphPen(m.ttglyphset)
                    layer.draw(pen)

                    m.ttglyphset._glyphs[g] = pen.glyph()

            except Exception as e:
                print(
                    "Problem converting glyph %s to quadratic. (Probably incompatible) "
                    % g
                )
                for m in f.masters:
                    m.ttglyphset._glyphs[g] = TTGlyphPen(m.ttglyphset).glyph()
            done[g] = True

        for g in exportable:
            do_a_glyph(g)

        fb.updateHead(
            fontRevision=f.version[0] + f.version[1] / 10 ** len(str(f.version[1])),
            created=timestampSinceEpoch(f.date.timestamp()),
            lowestRecPPEM=10,
        )
        fb.setupGlyf(f.default_master.ttglyphset._glyphs)
        fb.setupHorizontalHeader(
            ascent=int(f.default_master.ascender),
            descent=int(f.default_master.descender),
        )

        f.names.typographicSubfamily = f.default_master.name
        f.names.typographicFamily = f.names.familyName
        fb.setupNameTable(f.names.as_nametable_dict())

        fb.setupOS2(
            sTypoAscender=int(f.default_master.ascender),
            sTypoDescender=int(f.default_master.descender),
            sCapHeight=int(f.default_master.capHeight),
            sxHeight=int(f.default_master.xHeight),
        )

        if f.axes:
            model = f.variation_model()
            axis_map = {}
            variations = {}
            for g in exportable:
                variations[g] = self.calculate_a_gvar(f, model, g, metrics[g][0])

            for ax in f.axes:
                ax.name = ax.name.as_fonttools_dict
                axis_map[ax.tag] = ax
            for instance in f.instances:
                instance.location = {
                    k: axis_map[k].map_backward(v) for k, v in instance.location.items()
                }
            fb.setupFvar(f.axes, f.instances)

            fb.setupGvar(variations)
            fb.setupAvar(f.axes)
        # Move glyph categories to fontfeatures
        for g in f.glyphs.values():
            if g.exported:
                f.features.glyphclasses[g.name] = g.category

        build_all_features(f, fb.font)
        fb.setupPost()

        for table, field, value in f.customOpenTypeValues:
            setattr(fb.font[table], field, value)

        fb.font.save(self.filename)

        # Rename to production
        rename_map = {g.name: g.production_name or g.name for g in f.glyphs}
        if rename_map:
            font = TTFont(self.filename)
            font.setGlyphOrder([rename_map.get(n, n) for n in font.getGlyphOrder()])
            if "post" in font and font["post"].formatType == 2.0:
                font["post"].extraNames = []
                font["post"].compile(font)
            font.save(self.filename)

    def calculate_a_gvar(self, f, model, g, default_width):
        master_layer = f.default_master.get_glyph_layer(g)
        if not g in f.default_master.ttglyphset._glyphs:
            return None
        default_g = f.default_master.ttglyphset._glyphs[g]
        all_coords = []
        for m in f.masters:
            layer = m.get_glyph_layer(g)
            basecoords = GlyphCoordinates(m.ttglyphset._glyphs[g].coordinates)
            if m.ttglyphset._glyphs[g].isComposite():
                component_point = GlyphCoordinates(
                    [
                        (layer_comp.pos[0], layer_comp.pos[1])
                        for layer_comp in layer.components
                    ]
                )
                basecoords.extend(component_point)
            phantomcoords = GlyphCoordinates([(0, 0), (layer.width, 0), (0, 0), (0, 0)])
            basecoords.extend(phantomcoords)
            all_coords.append(basecoords)
        deltas = model.getDeltas(all_coords)
        gvar_entry = []
        if default_g.isComposite():
            endPts = list(range(len(default_g.components)))
        else:
            endPts = default_g.endPtsOfContours

        for delta, sup in zip(deltas, model.supports):
            if not sup:
                continue
            var = TupleVariation(sup, delta)
            # This assumes we do the default master first, which may not be true
            delta_opt = iup_delta_optimize(delta, deltas[0], endPts, tolerance=0.5)
            if None in delta_opt:
                var = TupleVariation(sup, delta_opt)
            gvar_entry.append(var)
        return gvar_entry

    # import numpy as np
    # def calculate_a_gvar(self, f, model, g, default_width):
    #     if not g in f.default_master.ttglyphset._glyphs:
    #         return None

    #     all_coords = []
    #     master_ix = f.masters.index(f.default_master)

    #     for m in f.masters:
    #         coords = list(m.ttglyphset._glyphs[g].coordinates)
    #         layer = m.get_glyph_layer(g)
    #         if m.ttglyphset._glyphs[g].isComposite():
    #             coords.extend([c.pos for c in layer.components])
    #         coords.extend( [ (0,0), (layer.width,0), (0,0), (0,0) ] )
    #         all_coords.append(np.array(coords))
    #     stacked = np.array(all_coords)
    #     defaults = stacked[np.newaxis,master_ix,:,:].repeat(len(f.masters),axis=0)
    #     base_deltas = stacked-defaults
    #     x_deltas = np.apply_along_axis(model.getDeltas, 1, base_deltas[:,:,0].transpose())
    #     y_deltas = np.apply_along_axis(model.getDeltas, 1, base_deltas[:,:,1].transpose())
    #     alldeltas = np.array([x_deltas, y_deltas]).transpose()
    #     gvar_entry = []
    #     for deltaset, sup in zip(alldeltas, model.supports):
    #         gvar_entry.append(TupleVariation(sup, list(map(tuple, deltaset))))
    #     return gvar_entry

class OpenType(TrueType):
    suffix = ".otf"

    def _load_layers(self, g):
        ttglyph = self.tt.getGlyphSet()[g]
        width = self.tt["hmtx"][g][0]
        layer = Layer(width=width, id=str(uuid.uuid1()) )
        layer._master = self.font.masters[0].id
        layer._font = self.font
        pen = RecordingPen()
        ttglyph.draw(pen)
        contours = pen.value
        lastcontour = []
        startPt = (0,0)
        lastPt = (0,0)
        index = 0
        for operation, segment in contours:
            if operation == "moveTo":
                startPt = segment[0]
            elif operation == "closePath":
                if startPt != lastPt:
                    lastcontour.append(Node(x=startPt[0], y=startPt[1],type = "l"))
                contour = Shape()
                contour.nodes = lastcontour
                layer.shapes.append(contour)
                lastcontour = []
            elif operation == "curveTo":
                lastcontour.append(Node(x=segment[0][0],y=segment[0][1],type = "o"))
                lastcontour.append(Node(x=segment[1][0],y=segment[1][1],type = "o"))
                lastcontour.append(Node(x=segment[2][0],y=segment[2][1],type = "c"))
                lastPt = segment[2]
            elif operation == "lineTo":
                lastcontour.append(Node(x=segment[0][0],y=segment[0][1],type = "l"))
                lastPt = segment[0]
            elif operation == "qCurveTo":
                lastcontour.append(Node(x=segment[0][0],y=segment[0][1],type = "o"))
                lastcontour.append(Node(x=segment[1][0],y=segment[1][1],type = "q"))

        return [layer]



from babelfont.convertors import Convert
import argparse
import sys

def main():
    parser = argparse.ArgumentParser(prog="babelfont", description='Convert between font formats')
    parser.add_argument('input', metavar='IN', help="Input file")
    parser.add_argument('output', metavar='OUT', help="Output file")
    args = parser.parse_args()

    try:
      font = Convert(args.input).load()
    except Exception as e:
      print("Couldn't read %s: %s" % (args.input, e))
      raise e
      sys.exit(1)

    try:
      Convert(args.output).save(font)
    except Exception as e:
      print("Couldn't write %s: %s" % (args.output, e))
      raise e
      sys.exit(1)

    sys.exit(0)

if __name__ == "__main__":
    sys.exit(main())



from dataclasses import dataclass
from .BaseObject import BaseObject
from fontTools.misc.transform import Transform
from .Node import Node
import math


@dataclass
class _ShapeFields:
    ref: str = None
    transform: Transform = None
    nodes: [Node] = None
    closed: bool = True
    direction: int = 1

    _layer = None

@dataclass
class Shape(BaseObject, _ShapeFields):
    @property
    def _write_one_line(self):
        return self.is_component

    @property
    def is_path(self):
        return not bool(self.ref)

    @property
    def is_component(self):
        return bool(self.ref)

    @property
    def component_layer(self):
        if not self.is_component:
            return None
        return self._layer.master.get_glyph_layer(self.ref)

    @property
    def pos(self):
        assert self.is_component
        if not self.transform:
            return (0,0)
        return tuple(self.transform[4:])

    @property
    def angle(self):
        assert self.is_component
        if not self.transform:
            return 0
        return math.atan2(self.transform[1], self.transform[0]) * 180 / math.pi

    @property
    def scale(self):
        assert self.is_component
        if not self.transform:
            return (1,1)
        print(self.transform)
        scaleX = math.sqrt(self.transform[0] ** 2 + self.transform[2] ** 2)
        scaleY = math.sqrt(self.transform[1] ** 2 + self.transform[3] ** 2)
        return (scaleX, scaleY)



from .BaseObject import BaseObject, I18NDictionary
from dataclasses import dataclass, asdict


@dataclass
class Names(BaseObject):
    """A table of global, localizable names for the font."""
    familyName: I18NDictionary = None
    designer: I18NDictionary = None
    designerURL: I18NDictionary = None
    manufacturer: I18NDictionary = None
    manufacturerURL: I18NDictionary = None
    license: I18NDictionary = None
    licenseURL: I18NDictionary = None
    version: I18NDictionary = None
    uniqueID: I18NDictionary = None
    description: I18NDictionary = None
    typographicFamily: I18NDictionary = None
    typographicSubfamily: I18NDictionary = None
    compatibleFullName: I18NDictionary = None
    sampleText: I18NDictionary = None
    WWSFamilyName: I18NDictionary = None
    WWSSubfamilyName: I18NDictionary = None
    copyright: I18NDictionary = None
    styleMapFamilyName: I18NDictionary = None
    trademark: I18NDictionary = None

    def __post_init__(self):
        for k in self.__dataclass_fields__.keys():
            if not getattr(self, k):
                setattr(self, k, I18NDictionary())

    def as_nametable_dict(self):
        rv = {}
        for k,v in asdict(self).items():
            if not v:
                continue
            rv[k] = v.default_or_dict()
        return rv



from dataclasses import dataclass
from .BaseObject import BaseObject


@dataclass
class Component(BaseObject):
    name: str
    position: list = None
    transform: list = None

    _serialize_slots = ["name", "transform"]



from babelfont.Font import Font
from babelfont.Axis import Axis
from babelfont.Glyph import Glyph
from babelfont.Master import Master
from babelfont.Instance import Instance
from babelfont.Guide import Guide
from babelfont.Anchor import Anchor
from babelfont.Layer import Layer
from babelfont.Shape import Shape
from babelfont.Node import Node
from babelfont.Names import Names
from babelfont.BaseObject import Color, Position, OTValue, I18NDictionary
from babelfont.convertors import Convert, BaseConvertor


def load(filename):
    return Convert(filename).load()



from datetime import datetime
from babelfont import *
import orjson
from fontTools.misc.transform import Transform
from babelfont.convertors import BaseConvertor
import uuid
import re


class Fontlab(BaseConvertor):
    suffix = ".vfj"

    def _load(self):
        self.fontlab = orjson.loads(open(self.filename).read())["font"]
        self._load_axes()
        self._load_masters()
        for g in self.fontlab.get("glyphs", []):
            glyph = self._load_thing(g, self.glyph_loader)
            self.font.glyphs.append(glyph)
        self.font.upm = self.fontlab.get("upm", 1000)
        return self.font

    axis_loader = (
        Axis,
        {
            "name": "name",
            "tag": "tag",
            "minimum": "min",
            "maximum": "max",
            "default": "default",
            "axisGraph": ("map", lambda _, x: [(v, int(k)) for k, v in x.items()]),
        },
    )

    def _load_kerning(self, flkerning):
        for kclass in flkerning.get("kerningClasses", []):
            self.font.features.namedClasses[kclass["name"]] = kclass["names"]
        kerning = {}
        for left, rvalue in flkerning.get("pairs", {}).items():
            for right, value in rvalue.items():
                kerning[(left, right)] = int(value)
        return kerning

    def _convert_color(_, col):
        r, g, b = int(col[1:3], 16), int(col[3:5], 16), int(col[5:7], 16)
        return Color(r, g, b, 0)

    def _convert_shapes(self, flshapes):
        shapes = []
        for shape in flshapes:
            if "component" in shape:
                shapes.append(
                    Shape(
                        ref=shape["component"]["glyphName"],
                        transform=Transform().translate(
                            shape.get("transform", {}).get("xOffset", 0),
                            shape.get("transform", {}).get("yOffset", 0),
                        ),
                        _=shape["component"],
                    )
                )
            else:
                for flcontour in shape["elementData"]["contours"]:
                    contour = Shape(nodes=[])
                    for n in flcontour["nodes"]:
                        contour.nodes.extend(self._load_node(n))
                    shapes.append(contour)
        return shapes

    def _load_node(self, input_):
        stuff = input_.split("  ")
        rv = []
        for ix, node in enumerate(stuff):
            m = re.match(r"(-?[\d\.]+) (-?[\d\.]+)( s)?", node)
            if not m:
                raise ValueError("Can't understand nodestring %s" % node)
            if len(stuff) == 1:
                ntyp = "l"
            elif len(stuff) == 3 and ix == 2:
                ntyp = "c"
                if m[3]:
                    ntyp = ntyp + "s"
            elif len(stuff) == 2 and ix == 1:
                ntyp = "q"
                if m[3]:
                    ntyp = ntyp + "s"
            else:
                ntyp = "o"
            rv.append(Node(x=float(m[1]), y=float(m[2]), type=ntyp))
        return rv

    master_loader = (
        Master,
        {
            "name": ["id", "name"],
            "kerning": ("kerning", _load_kerning),
            "location": (
                "location",
                lambda s, x: {s.axis_name_map[k].tag: v for k, v in x.items()},
            ),
        },
    )

    _layer_loader = (
        Layer,
        {
            "advanceWidth": "width",
            "name": ["name", "_master"],
            "color": ("color", _convert_color),
            "elements": ("shapes", _convert_shapes),
        },
    )

    glyph_loader = (
        Glyph,
        {
            "name": "name",
            "unicode": (
                "codepoints",
                lambda _, x: [int(cp, 16) for cp in x.split(",")],
            ),
            "layers": ("layers", _layer_loader),
            "elements": None,
        },
    )

    def _load_thing(self, thing, handler):
        kwargs = {}
        kwargs["_"] = {}
        cls, mapping = handler
        for k, v in thing.items():
            if k not in mapping:
                kwargs["_"][k] = v
            elif mapping[k] is None:
                continue
            elif isinstance(mapping[k], list):
                for newname in mapping[k]:
                    kwargs[newname] = v
            elif isinstance(mapping[k], str):
                kwargs[mapping[k]] = v
            elif isinstance(mapping[k], tuple) and callable(mapping[k][1]):
                newName, convertor = mapping[k]
                kwargs[newName] = convertor(self, v)
            else:
                newName, convertor = mapping[k]
                if isinstance(v, list):
                    kwargs[newName] = [self._load_thing(v1, convertor) for v1 in v]
                else:
                    kwargs[newName] = self._load_thing(v, convertor)
        obj = cls(**kwargs)
        obj._font = self.font
        return obj

    def _load_axes(self):
        self.axis_name_map = {}
        for ax in self.fontlab.get("axes", []):
            axis = self._load_thing(ax, self.axis_loader)
            self.axis_name_map[ax["shortName"]] = axis
            self.font.axes.append(axis)

    def _load_masters(self):
        for m in self.fontlab.get("masters", []):
            fl_master = m["fontMaster"]
            metric_dict = {}
            for metric in [
                "ascender",
                "descender",
                "xHeight",
                "capsHeight",
                "measurement",
                "underlineThickness",
                "underlinePosition",
            ]:
                if metric not in fl_master:
                    continue
                value = fl_master[metric]
                del fl_master[metric]
                if metric == "capsHeight":
                    metric = "capHeight"
                metric_dict[metric] = value
            master = self._load_thing(fl_master, self.master_loader)
            master.metrics = metric_dict
            master.font = self.font
            self.font.masters.append(master)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['babelfont', 'babelfont.convertors', 'babelfont.fontFilters']

package_data = \
{'': ['*']}

install_requires = \
['cu2qu>=1.6.7,<2.0.0',
 'fontfeatures>=1.0.6,<2.0.0',
 'fonttools>=4.21.1,<5.0.0',
 'glyphsLib>=5.3.2',
 'openstep-plist>=0.2.2',
 'orjson>=3.5.1,<4.0.0',
 'ufoLib2>=0.11.1,<0.12.0']

entry_points = \
{'console_scripts': ['babelfont = babelfont.__main__:main']}

setup_kwargs = {
    'name': 'babelfont',
    'version': '3.0.0a7',
    'description': '',
    'long_description': None,
    'author': 'Simon Cozens',
    'author_email': 'simon@simon-cozens.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
}


setup(**setup_kwargs)



from fontFeatures import Attachment, Routine, Positioning, ValueRecord
import logging


log = logging.getLogger(__name__)


def build_all_features(font, ttFont):
    build_cursive(font)
    build_mark_mkmk(font)
    build_mark_mkmk(font, "mkmk")
    build_kern(font)

    font.features.buildBinaryFeatures(ttFont, font.axes)


def build_kern(font):
    def _expand_class(c):
        if c[0] != "@":
            return [c]
        if c[1:] not in font.features.namedClasses:
            logging.info("Attempted to use undefined kerning class %s" % c)
            return None
        return [x for x in font.features.namedClasses[c[1:]] if font.glyphs[x].exported]

    kernroutine = Routine()
    for (l,r), kern in font._all_kerning.items():
        l, r = _expand_class(l), _expand_class(r)
        if l is None or r is None:
            continue
        kernroutine.rules.append(Positioning(
            [ l, r ],
            [ ValueRecord(xAdvance=kern), ValueRecord() ],
        ))
    font.features.addFeature("kern", [kernroutine])


def build_cursive(font):
    anchors = font._all_anchors
    if "entry" in anchors and "exit" in anchors:
        attach = Attachment(
            "entry", "exit", anchors["entry"], anchors["exit"],
            flags=(0x8 | 0x1)
        )
        r = Routine(rules=[attach], )
        font.features.addFeature("curs", [r])

def build_mark_mkmk(font, which="mark", strict=False):
    # Find matching pairs of foo/_foo anchors
    anchors = font._all_anchors
    r = Routine(rules=[])
    if which == "mark":
        basecategory = "base"
    else:
        basecategory = "mark"
    for baseanchor in anchors:
        markanchor = "_" + baseanchor
        if markanchor not in anchors:
            continue
        # Filter glyphs to those which are baseanchors
        bases = {
            k: v
            for k, v in anchors[baseanchor].items()
            if font.glyphs[k].exported and (font.glyphs[k].category == basecategory)
        }
        marks = {
            k: v
            for k, v in anchors[markanchor].items()
            if font.glyphs[k].exported and (not strict or font.glyphs[k].category == "mark")
        }
        if not (bases and marks):
            continue
        attach = Attachment(baseanchor, markanchor, bases, marks)
        attach.fontfeatures = font.features  # THIS IS A TERRIBLE HACK
        r.rules.append(attach)
    if r.rules:
        font.features.addFeature(which, [r])
